@@ -3,7 +3,7 @@
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 // Mobius Forensic Toolkit
-// Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021,2022 Eduardo Aguiar
+// Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021,2022,2023,2024 Eduardo Aguiar
 //
 // This program is free software; you can redistribute it and/or modify it
 // under the terms of the GNU General Public License as published by the
@@ -27,10 +27,9 @@
 #include <cstdint>
 #include <string>
 #include <stdexcept>
+#include <type_traits>
 
-namespace mobius
-{
-namespace py
+namespace mobius::py
 {
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 // Parse function arguments
@@ -64,6 +63,22 @@ std::uint16_t get_arg_as_uint16_t (PyObject *, std::uint32_t, std::uint16_t);
 std::uint32_t get_arg_as_uint32_t (PyObject *, std::uint32_t, std::uint32_t);
 std::uint64_t get_arg_as_uint64_t (PyObject *, std::uint32_t, std::uint64_t);
 
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief Check argument type
+//! \param args Function arguments
+//! \param idx Argument number, starting in 0
+//! \param f Check function
+//! \return true/false
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+template <typename F> bool
+check_arg (PyObject *args, std::uint32_t idx, F f)
+{
+  if (idx < get_arg_size (args))
+    return f (get_arg (args, idx));
+
+  throw std::out_of_range (MOBIUS_EXCEPTION_MSG ("argument index out of range"));
+}
+
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 //! \brief Get function argument and convert using a function
 //! \param args Function arguments
@@ -101,6 +116,8 @@ void set_invalid_type_error (const std::string&);
 void set_value_error (const std::string&);
 void set_io_error (const std::string&);
 void set_runtime_error (const std::string&);
+void set_stop_iteration ();
+std::string get_error_message ();
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 // Python reflection
@@ -110,10 +127,17 @@ std::string get_filename ();
 std::string get_funcname ();
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-// C++ <-> Python type conversion
+// Singleton types
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+PyObject *pynone ();
+PyObject *py_false ();
+PyObject *py_true ();
+PyObject *py_not_implemented ();
+
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+// C++ <-> Python concrete types conversion
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 bool pynone_check (PyObject *);
-PyObject *pynone ();
 
 bool pystring_check (PyObject *);
 PyObject *pystring_from_std_string (const std::string&);
@@ -158,6 +182,58 @@ bool pyfloat_check (PyObject *);
 double pyfloat_as_cpp (PyObject *);
 PyObject *pyfloat_from_cpp (double);
 
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+// C++ <-> Python type conversion
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+template <typename T>
+bool
+isinstance (PyObject *value, T *type)
+{
+  return PyObject_IsInstance (value, reinterpret_cast<PyObject *> (type));
+}
+
+template <typename O, typename T>
+auto
+from_pyobject (PyObject *value, T *type)
+{
+  if (isinstance (value, type))
+    return * (reinterpret_cast <O *>(value)->obj);
+
+  throw std::invalid_argument (mobius::MOBIUS_EXCEPTION_MSG ("object must be an instance of " + std::string (type->tp_name)));
+}
+
+template <typename O, typename Tcpp, typename T>
+PyObject *
+to_pyobject (const Tcpp& obj, T *type)
+{
+  PyObject *ret = _PyObject_New (type);
+
+  if (ret)
+    ((O *) ret)->obj = new Tcpp (obj);
+
+  return ret;
+}
+
+template <typename O, typename Tcpp, typename T>
+PyObject *
+to_pyobject_nullable (const Tcpp& obj, T *type)
+{
+  PyObject *ret = nullptr;
+
+  if (obj)
+    {
+      ret = _PyObject_New (type);
+
+      if (ret)
+        ((O *) ret)->obj = new Tcpp (obj);
+    }
+
+  else
+    ret = mobius::py::pynone ();
+
+  return ret;
+}
+
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 //! \brief Check setter value
 //! \param value Value object
@@ -201,38 +277,7 @@ pytuple_from_cpp_pair (const T& p, F1 pyf1, F2 pyf2)
       PyTuple_SetItem (ret, 1, pyf2 (p.second));
     }
 
-  return ret;  
-}
-
-// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-//! \brief Create new Python list from C++ container
-//! \param container C++ container
-//! \param pyfunc Function to convert C++ items to Python objects
-//! \return Python list
-// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-template <typename C, typename F> PyObject *
-pylist_from_cpp_container (const C& container, F pyfunc)
-{
-  PyObject *ret = PyList_New (0);
-
-  if (!ret)
-    return nullptr;
-
-  for (const auto& item : container)
-    {
-      PyObject *py_item = pyfunc (item);
-
-      if (!py_item)
-        {
-          Py_CLEAR (ret);
-          return nullptr;
-        }
-
-      PyList_Append (ret, py_item);
-      Py_DECREF (py_item);
-    }
-
-  return ret;  
+  return ret;
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -277,12 +322,6 @@ pydict_from_cpp_container (const C& container, FK pyfk, FV pyfv)
   return ret;
 }
 
-} // namespace py
-} // namespace mobius
-
-// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-// Deprecated (must use mobius::py::pystring...)
-// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-PyObject *PyString2_from_std_string (const std::string&);
+} // namespace mobius::py
 
 #endif
